// This may look like C code, but it is really -*- C++ -*-
/*
 ************************************************************************
 *
 *		HyperText Transport Protocol (HTTP) support
 *
 * This code implements the basic flow of a HTTP transaction, as defined in
 * a HTTP 1.1 document [RFC 2068]. That is, this code performs:
 * 	- opening of an HTTP connection (directly or via a proxy),
 * 	- sending of a request,
 *	- listening to a reply,
 *	- analyzing of the return code,
 *	- parsing of the response headers,
 *	- dispatching to handle reply's data,
 *	- closing of the connection.
 * 
 * All specific processing (sending of a request content of a particular type,
 * processing of a reply content) is supposed to be done in user-specialized
 * classes (interfaces) RequestContent and ContentConsumer.
 * 
 * In particular, when an HTTPTransaction object establishes a connection, it
 * sends a request line followed by generic request headers Host:,
 * User-Agent: and (if enabled) Authorization:. A specialized RequestContent
 * object would be called then to write other request headers and the request
 * content if needed.
 *
 * $Id: HTTP.cc,v 1.5 1998/09/14 18:51:26 oleg Exp oleg $
 *
 ************************************************************************
 */

#if defined(__GNUC__)
#pragma implementation
#endif

#include "HTTP.h"
#include "MIMEEntity.h"
#include <fstream.h>


		// Perform a single interaction with the server:
		// open a connection, send a request,
		// listen to a reply, handle it, and close the connection
HTTPTransaction::TransStatus HTTPTransaction::perform_transaction
	(RequestContent& request_content,
	 ContentConsumer& content_consumer)
{
  Logger clog;
  
  //ios::binary
  TCPstream link;
  //(link.rdbuf())->dump("before setbuf");
  
  clog << "\nConnecting to " << connection_parms.q_host_to_connect() << ':' 
      << connection_parms.q_port_to_connect() << endl;
      
  link.connect(SocketAddr(connection_parms.q_host_to_connect(),
  			  connection_parms.q_port_to_connect()));
  
  if( !link.is_open() || !link )
    return clog << "\nFailed to establish the connection", Error;
   

			// The following if() shouldn't be necessary, as
			// any HTTP server ought to accept the full URL
			// (according to HTTP 1.1)
			// Alas, there are many servers out there that return
			// '404 Not found' for the full URL
			// Therefore, if we connect to a server directly
			// (not via a proxy), we POST only the resource
			// path of the URL
  if( !connection_parms.q_use_proxy() )
    link << request_content.q_method_str() << ' ' 
         << connection_parms.resource_path
	 << " HTTP/1.0" << CRLF;
  else
    link << request_content.q_method_str() << " http://"
         << connection_parms.server_name << ':'
         << connection_parms.server_port << connection_parms.resource_path
         << " HTTP/1.0" << CRLF;
  link << "Host: " << connection_parms.server_name << ':'
       << connection_parms.server_port << CRLF;
  link << "User-Agent: " << connection_parms.identity_str << CRLF;
  if( connection_parms.q_send_auth() )
    link << "Authorization: " << connection_parms.auth_credential << CRLF;

  link << request_content;
  link.flush();
  
  if( !link )
    return clog << "\nError sending request", Error;
   //(link.rdbuf())->dump("after sent");
 
  clog << "sent request. Now listening for the response..." << endl;
  
  return process_response(link,content_consumer);
}


			// Read the first line of the server's
			// response, something like
			//  HTTP/1.x 200 OK
			// and extract the response code
			// Return EOF on error
			// All activity (and errors) are logged into the
			// clog stream
int HTTPTransaction::read_response_status_line(istream& http_stream)
{
  Logger clog;
  char line_buffer[81];
  if( ! http_stream.get(line_buffer,sizeof(line_buffer)-1,'\r').good() )
    return (clog << "error reading a response line from the HTTP stream"), EOF;
    
  clog << "Got response: >" << line_buffer;
			// dump the rest of the response line
			// (if was longer than sizeof(line_buffer))
  for(char c; http_stream.get(c).good() && c != '\n'; )
    if( c != '\r' )
      clog << c;
  clog << "<" << endl;

  int ret_code = EOF;
  if( sscanf(line_buffer,"HTTP/1.%*d %3d",&ret_code) != 1 )
    return (clog << "invalid response's status line (quoted above)"), EOF;
  return ret_code;
}

				// Receive HTTP server's response
				// and process it
HTTPTransaction::TransStatus HTTPTransaction::process_response(istream& link,
	ContentConsumer& content_consumer)
{
  Logger clog;
  const int resp_code = read_response_status_line(link);
  clog << "\nreceived response code " << resp_code << endl;
  if( resp_code == EOF )
    return ReTry;
  RFC822Headers headers(link);		// read the headers
  clog << headers;			// and dump them
  
  if( resp_code == 200 )
    return content_consumer(link,headers) ? OK :
 		(clog << "Error handling content", Error);
  else if( resp_code == 304 )
    return content_consumer.handle_no_change(headers) ? OK : Error;
  return content_consumer.handle_other_response(resp_code,link,headers);
}


bool ContentConsumer::handle_no_change(const RFC822Headers& headers)
{
   Logger() << "The server has nothing new";
   return true;
}

HTTPTransaction::TransStatus
ContentConsumer::handle_other_response(const int resp_code,
			      istream& link, const RFC822Headers& headers)
{
  switch(resp_code)
  {
    case 200:				// OK, result follows
         return HTTPTransaction::Error;	// somehow it wasn't handled!
  	 
    case 201:				// Created. No content follows
         clog << "The server has created the resource." << endl;
         return HTTPTransaction::OK;
           
    case 202:				// Accepted, check for the result later
         clog << "The request was accepted, but not yet fulfilled; check later"
              << endl;
  	 return HTTPTransaction::ReTry;
  	 
    case 304:				// Nothing new
        return HTTPTransaction::Error;	// somehow it wasn't handled!
  	 
    default:
  	 clog << "Got apparently wrong/error respose. The rest of it follows\n"
  	      << link.rdbuf() << endl;
  	 return HTTPTransaction::Error;
  }
  return HTTPTransaction::Error;
}

TypedFile::TypedFile(const char _file_name [], const char _file_type [])
{
  assert( strlen(_file_name) < sizeof(file_name) - 1 );
  strcpy(const_cast<char *>(&file_name[0]),_file_name);
  assert( strlen(_file_type) < sizeof(type) - 1 );
  strcpy(const_cast<char *>(&type[0]),_file_type);
}

		// Send an _attributed_ file content down the link.
		// That is, we send the corresponding Content-
		// headers, followed by CRLF and the file data
		// themselves.
ostream& operator << (ostream& link, const TypedFile& file)
{
  ifstream req_file(file.file_name,ios::in|ios::binary);
  if( !req_file )
    _error("Failed to open the request file '%s'",file.file_name);
  
  link << "Content-Type: " << file.type << CRLF;
  link << "Content-Length: " << get_file_size(file.file_name) << CRLF;
  link << CRLF;			// an empty line - end of headers
  
  assert( link.good() );
  
  link << req_file.rdbuf();
  
  return link;
}


const char * RequestContent::q_method_str(void) const
{
  static const char * method_str [] = { 
  	"INVALID", "GET", "POST", "PUT", "HEAD", "OPTIONS", "DELETE" };
  assert( static_cast<unsigned int>(method) < 
	  sizeof(method_str)/sizeof(method_str[0]) );
  return method_str[static_cast<unsigned int>(method)];
}

