// This may look like C code, but it is really -*- C++ -*-
/*
 ************************************************************************
 *
 * 			Handling of MIME Entities
 *
 * According to RFC 2045, "Multipurpose Internet Mail Extensions (MIME)
 * Part One, Format of Internet Message Bodies",
 *
 * "The term 'entity', refers specifically to the MIME-defined header
 * fields and contents of either a message or one of the parts in the
 * body of a multipart entity.  The specification of such entities is
 * the essence of MIME.  Since the contents of an entity are often
 * called the 'body', it makes sense to speak about the body of an
 * entity.  Any sort of field may be present in the header of an entity,
 * but only those fields whose names begin with "content-" actually have
 * any MIME-related meaning."
 *
 * Specifically, the MIME standard (RFC 2045) defines the following
 * MIME-related headers (header fields)
 *	Content-type
 *	Content-Transfer-Encoding
 *	Content-ID
 *	Content-Description
 *
 * See the RFC for more details. At present, we ignore Content-ID,
 * Content-Description and Content-Transfer-Encoding headers (thus
 * assuming that entity's body is being transferred as it is), and
 * we _demand_ a Content-type header (thus, its absense is considered
 * an error). Generally we leave content interpretation and processing
 * to Mailcap (see Mailcap.{h,cc}) or some other user-supplied handler.
 * However, if the MIME entity turns out to be composite (multipart),
 * we disassemble it into separate discrete parts, and have them
 * handled, in turn. Composite entities are distinguished by their
 * Content-type (media type) of multipart/mixed, multipart/alternative,
 * multipart/parallel, multipart/digest, or some other multipart type.
 * At present, all of them are handled the same way.
 *
 * $Id: MIMEEntity.h,v 4.1 1997/08/06 18:24:53 oleg Exp oleg $
 *
 ************************************************************************
 */

#ifndef __GNUC__
#pragma once
#else
#pragma interface
#endif

#ifndef __MIMEEntity_h
#define __MIMEEntity_h

#include "util.h"

			// RFC822 (and HTTP) character types
			// Section "2.2 Basic Rules" of the HTTP 1.1
			// document (see the reference above)
class HTTPCharType
{
  enum { _token=1 };
  static char char_types [128];
public:
  enum { UPALPHA, LOALPHA, ALPHA, DIGIT, CTL, CR, LF, SP, HT, QU,
  HEX, TOKEN_CHAR, TSPECIAL };
  static bool is_token_char(const unsigned char c)
  	{ return c < 128 && (char_types[c] & _token); }
  static bool is_lws_char(const char c) { return c == ' ' || c == '\t'; }
};

		// A collection of RFC822 message headers ("Headers")
		// section 4.2 of the HTTP 1.1 document
		// An RFC822Headers object holds headers arranged in the
		// reverse order (of their appearance in the input stream
		// at the time the object was constructed)
		// Thus if there are two headers with the same
		// name, the last one takes precedence (would
		// be found during lookup)
class RFC822Headers : public Dictionary
{
  RFC822Headers(const RFC822Headers&);	 // private and unimplemented to
  void operator = (const RFC822Headers&);// prohibit copy-constructing
  
  enum Digesting_Result { Continue, Empty_Line, IO_Error, Syntax_Error };
  Digesting_Result read_a_header(istream& in_stream);
  
public:
  				// Load the headers off the stream
  RFC822Headers(istream& in_stream);
 
  friend ostream& operator << (ostream& os, const RFC822Headers& headers);
};

			// A particular dictionary holding parsed
			// string of something like
			// extended-token *( ";" parameter )
			// (for example, Content-Type: header has
			// such a structure)
class ParameterizedToken : public Dictionary
{
  ParameterizedToken(const ParameterizedToken&);// private and unimplemented to
  void operator = (const ParameterizedToken&);	// prohibit copy-constructing
  
public:
  				// Load the dictionary
  ParameterizedToken(const char first_key_name [], iststream& source);
 
//  friend ostream& operator << (ostream& os, const ParameterizedToken& headers);
};

class ContentType : public ParameterizedToken
{
  ContentType(const ContentType&);	// private and unimplemented to
  void operator = (const ContentType&);	// prohibit copy-constructing

  static const char MediaType_Key [];

public:
  ContentType(iststream& source)
  	: ParameterizedToken(MediaType_Key,source) {}
  
  bool q_multipart(void) const;
  const StringCollection::Mark q_media_type_mark() const
  	{ return lookup(MediaType_Key); }
};

	// The following MIMEEntity class is to represent an
	// _elementary_ (discrete, not multipart!) MIME entity in an already
	// "internalized" form, with the headers put into a dictionary, and
	// a body loaded into a specified file (properly decoded, assembled,
	// fetched, etc)

	// The purpose of this class is to load the entity's body, and
	// to hold parameters for the EntityHandler;
	// The user is never supposed to directly create an object
	// of this class (that's why the class got no public constructors)
	// Besides, the object contains a few references to external
	// (hopefully, pre-allocated) data: use with care!
class MIMEDiscreteEntity
{
  friend class MIMEEntityStream;

  const RFC822Headers& headers;
  const ContentType& content_type;
  char body_file_name [L_tmpnam];	// Where the processed body is stored
  bool valid_flag;
  
  MIMEDiscreteEntity(const MIMEDiscreteEntity&); // private and unimplemented to
  void operator = (const MIMEDiscreteEntity&);   // prohibit copy-constructing
  
  				// Load the entity's body of a given stream
  				// (terminating at boundary if boundary is
  				// not an empty string)
  				// If boundary is given it will be scanned
  				// completely when MIMEDiscreteEntity
  				// successfully completes
  MIMEDiscreteEntity(const RFC822Headers& _headers,
  		     const ContentType& _content_type,
  		     istream& en_stream, const char boundary[]);
  ~MIMEDiscreteEntity(void);
  
  void leave_content_file(void);	// Do not remove the content file
  					// upon destruction

   
  				// Receive the content data into
  				// the given content_stream
  				// The functions below differ in detecting of
  				// input termination. Reading of the input
  				// stream may be terminated by EOF, scanning
  				// of given number of bytes, or encountering
  				// a given boundary
  				
  static bool receive_content_data(ostream& content_file, istream& en_stream);
  static bool receive_content_data(ostream& content_file, istream& en_stream,
  						   const int length);
  static bool receive_content_data(ostream& content_file, istream& en_stream,
  						   const char boundary[]);
  static bool receive_content_data(ostream& content_file, istream& en_stream,
  				const int length, const char boundary[]);

public:
  bool is_valid(void) const		{ return valid_flag; }
  const char * q_body_file_name(void) const
  		{ assert(valid_flag); return body_file_name; }
  const RFC822Headers& q_headers(void) const	{ return headers; }
  const ContentType& q_content_type(void) const	{ return content_type; }
};

		// A generic EntityHandler
		// It is given a complete discrete MIME entity, which the handler
		// (operator ()) may do anything with.
		// The handler should return the result how successful it
		// was in handling the entity
struct EntityHandler
{
  virtual bool operator () (const MIMEDiscreteEntity& entity) = 0;
};

		// The following MIMEEntityStream class represents
		// a (discrete or composite) MIME entity in an externalized
		// form, that is, in the form the entity is sent
		// over communication channels. The entity's data
		// may be encoded/transformed for transmission.
		// The entity can be composite (multipart), in which case
		// the stream possibly encodes several DiscreteEntities,
		// which may need to be handled separately.
		// This externalized format is the subject of
		// RFC 2045
		//
		// The MIMEEntityStream stream actually provides a
		// namespace for a host of functions dealing with the
		// stream. The stream object itself is not available
		// to the user, because it is supposed to be transient.
		// The only way the user can use this object is to
		// call static methods (factory methods) handle_entity()
		// and pass an entity consumer (that is, a handler)
		// The entity would be loaded off the stream, internalized,
		// and passed to a handler. If the entity turns out to be
		// composite, it would be disassembled; in that case, the
		// handler would be called several times.

class MIMEEntityStream
{
  istream& en_stream;
  const RFC822Headers& headers;
  const char * const boundary;
  
  MIMEEntityStream(const MIMEEntityStream&); // private and unimplemented to
  void operator = (const MIMEEntityStream&); // prohibit copy-constructing
  
protected:
			// The object itself is constructed and destroyed behind
			// the scenes. Note, two of its members (headers
			// and boundary) are _references_ to exterenal
			// data. Whoever creates the object must make sure
			// the object would be destroyed _before_ these
			// data. Fortunately, it's our static members
			// handle_entity() who are allowed to construct
			// this MIMEEntityStream object; hope they would
			// take care of this reference constraint

				// Entity's headers are supplied separately
  				// The stream contains only entity's body
  MIMEEntityStream(const RFC822Headers& headers, istream& body_stream,
  			const char _boundary[] = "" );
  
  bool operator >> (const EntityHandler& handler);
  
  bool eof(void) const		{ return en_stream.eof(); }
  
  			// Check to see if the current symbol is
  			// CR, LF, or CR followed by LF
  			// If so, scan through it and return true
  bool seen_CRLF(void);
  
  			// Check to see if there are two dashes at
  			// the current position in the stream.
  			// If so, scan them and return true.
  			// Otherwise, the current position in the
  			// stream remains unchanged
  bool seen_dash_dash(void);
  
  			// Skip linear-white spaces (space or HT)
  			// The stream will be positioned at the first
  			// not-LWS character, or at EOF
  void skip_lwspaces(void);
  
  			// Skip the stream all the way through
  			// CRLF -- boundary
  			// Return true if succeeded, in that case
  			// the stream is positioned right after the
  			// last character of the boundary
  			// Return false if EOF or other error occurred
  bool skip_through_dash_boundary(const char boundary []);
  
public:
	
			// These are the only public "members" of the
			// MIMEEntityStream stream class
			// They take a stream containing an entity
			// (or entity's body) and an entity's consumer
			// handler. They return true only if the entity
			// was loaded _and_ handled successfully.
			// In that case, the stream would be positioned
			// right after the last character of entity's
			// body.

 				// Entity's headers are supplied separately
  				// The stream contains only entity's body
  static bool handle_entity(const RFC822Headers& headers,
  		     istream& in_stream, const EntityHandler& handler)
  	{ return MIMEEntityStream(headers,in_stream) >> handler; }

	
				// The in_stream contains the whole entity
  static bool handle_entity(istream& in_stream, const EntityHandler& handler)
  	{ return handle_entity(RFC822Headers(in_stream),in_stream,handler); }
 
};

#endif
