// This may look like C code, but it is really -*- C++ -*-
/*
 ************************************************************************
 *
 *			  HTTPFS Client
 *
 * This is an interface to  a HTTPFS client framework, which communicates
 * with a HTTPFS server and maintains a local cache of fetched files and
 * directories (VNodes). This interfaces declares VNodes and VNode_list --
 * which constitute a public HTTPFS client interface, -- along with
 * a few utility classes.
 * 	
 * $Id: mchttp-client.h,v 1.5 1999/05/02 22:48:33 oleg Exp oleg $
 *
 ************************************************************************
 */

#if !defined(__GNUC__)
#pragma once
#else
#pragma interface
#endif

#ifndef __MCHTTP_CLIENT_H
#define __MCHTTP_CLIENT_H

#include "HTTP.h"
#include "myenv.h"

#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif

		// Defines the size of the buffers that hold the path names:
		// Max number of characters in a (virtual) path name
#define VPATH_MAX 255	


class RFC822Headers;

	// Parse a VFS file name into HTTP parameters
	// The VFS file name is expected as:
	// /hostname[:port][^proxyname[:proxyport]]/resource-path
	// where hostname and proxyname are either names or IP addresses
	// resource path is typically
	//	/cgi-bin/admin/MCHFS-server.pl/images/
	// or
	//	/cgi-bin/admin/MCHFS-server.pl/DeepestRoot/tmp/t

class VFSNameParser : public HTTPConnection::CTOR
{
  const char * const vfs_name;
  void operator=(const VFSNameParser&);
  VFSNameParser(const VFSNameParser&);	// cloning is not allowed
  VFSNameParser(const char _vfs_name []) : vfs_name(_vfs_name) {}
public:
		// No public constructors! The object is transient
  static VFSNameParser vfs_name_to_httpfs(const char _vfs_name [])
  { return VFSNameParser(_vfs_name); }
  void fill_in(HTTPConnection& connection) const;
};

		// A C++ wrapper around a familiar struct stat
		// A FileStatus considered invalid if it is zeroed out
		// (in particular, if the st_atime field is zero)
class FileStatus : public stat
{
 bool load_from_status_info(const char status_info []);
 bool load_from_etag(const RFC822Headers& headers);
public:
   		// check to see if an object is valid
 operator bool (void) const { return st_atime > 0; }  
 void invalidate(void)		{ st_atime = 0; }
 FileStatus(void) { invalidate(); }
 FileStatus(const char status_info []) { load_from_status_info(status_info); }
 FileStatus(const RFC822Headers& headers) { load_from_etag(headers); }
};

struct RequestWithCommand : public RequestContent
{
  const char * const cmd;
  void send_request_data (ostream& link)
  { link << "Pragma: httpfs=\"" << cmd << '"' << CRLF;
    RequestContent::send_request_data(link);
  }
public:
 RequestWithCommand(RequestContent::METHOD method,
		    const char httpfs_cmd []) : 
  RequestContent(method), cmd(httpfs_cmd) {}
};

struct RequestContent_IfModif : public RequestContent
{
  const time_t modif_time;	// could be zero, in which case the
  				// request is unconditional
  virtual void send_request_data (ostream& link);
public:
 RequestContent_IfModif(RequestContent::METHOD method,
		    const time_t _modif_time) : 
  RequestContent(method), modif_time(_modif_time) {}
};

/*
 *-------------------------------------------------------------------------
 *			Vnodes and V-node cache
 * These data structures are intended to hold all the information we
 * have received from a server (concerning some particular file or directory).
 * Should this file or directory be requested again, we merely check
 * with the server if the cached information is still valid.
 */

class VNode
{
  char name[VPATH_MAX+1];		// (VFS) name of this object
  FileStatus stat_info;			// stat of this particular object
  time_t last_checked;			// When was the last time the object
  					// or its status was obtained
  char cache_file[L_tmpnam+1];		// Name of the file that caches this
  					// object's content; empty name means
  					// the object is not cached
  int cache_handle;			// If this VNode and the cache_file
  					// is open, this is the file handle
  int ref_count;
  
  VNode(const VNode&);			// Unimplemented and forbidden: cloning
  void operator = (const VNode&);	// is not allowed
protected:
			// Take the content HTTP server sends us
 			// and load it into cache
  struct CacheLoad : public ContentConsumer
  {
    VNode& vnode;
    bool operator () (istream& link, const RFC822Headers& headers);
    HTTPTransaction::TransStatus
    	handle_other_response(const int resp_code,
			      istream& link, const RFC822Headers& headers);
    CacheLoad(VNode& _vnode) : vnode(_vnode) {}
  };
  friend class CacheLoad;
			// Refresh the status_info after the server
 			// has created a resource
  struct PUTRefresh : public ContentConsumer
  {
    VNode& vnode;
    bool operator () (istream& link, const RFC822Headers& headers)
    	{ return false; }
    HTTPTransaction::TransStatus
    	handle_other_response(const int resp_code,
			      istream& link, const RFC822Headers& headers);
    PUTRefresh(VNode& _vnode) : vnode(_vnode) {}
  };
  friend class PUTRefresh;
public:

 		// At any given time, a VNode can be either:
 		//   - invalid
 		//   - open (in use), by exactly one client
 		//   - unreferenced (but still holding valid cache data)
 		// As we have a place for only one cache_handle,
 		// ref_count field may be at most one
 
   		// check to see if an object is valid
  operator bool (void) const	{ return (bool)stat_info && last_checked > 0; }
  bool q_in_use(void) const	{ return ref_count > 0; }
  void invalidate(void)
  		{ stat_info.invalidate(); last_checked = 0; 
	          if( cache_handle >= 0 ) ::close(cache_handle);
		  cache_handle = -1; }

  VNode(const char name []);
  ~VNode(void);
  bool match(const char another_name []) const 
  		{ return strcmp(name,another_name) == 0; }
  virtual bool open(const int open_flags);
  bool open_cache_handle(const int open_flags);
  virtual bool close(void);
  //int get_ref_count(void) const		{ return ref_count; }
  const char * get_name(void) const	{ return name; }
  const char * get_cache_file(void) const	{ return cache_file; }
  int get_cache_handle(void) const	{ return cache_handle; }
  FileStatus get_stat(void) const	{ return stat_info; }
  time_t get_last_checked(void) const	{ return last_checked; }
  
  			// Make sure the VNode is in a logically consistent
  			// state
  void assert_consistency(void) const;
};


			// The list (and cache) of VNodes
			// The 1-based index in this list is vnode's handle
			// Note, after a VNode is closed, it's not automatically
			// disposed of. For one thing, the vnode maybe in use.
			// For another, even if the vnode is currently
			// unreferenced, it may be needed again in the future.
class VHandle
{
  friend class VNode_list;
  
  int handle;			// must be positive!

public:
  VHandle(const int _handle) : handle(_handle) {}
  VHandle(void) : handle(-1) {}
  operator bool(void) const		{ return handle > 0; }
  int get_raw_handle(void) const	{ return handle; }
  friend ostream& operator << (ostream& os, const VHandle vhandle)
  { return (bool)vhandle ? (os << "vnode#" << vhandle.handle) :
    os << "invalid vhandle"; }
};


class VNode_list
{
  const size_t size;		// The size of the list
  VNode ** const vnode_array;

  struct VListPtr { 
    VHandle handle; VNode ** vpp;
    VListPtr(void) : vpp(0) {}
    VListPtr(const int _handle, VNode ** _vpp) :  handle(_handle), vpp(_vpp) {}
    operator bool(void) const	{ return (bool)handle && vpp != 0 && *vpp != 0; }
    VListPtr(const VNode_list& list, const char name []); // lookup by name
    VListPtr(const VNode_list& list, const VHandle handle); // lookup by handle
    VListPtr(const VNode_list& list, const int cache_handle); // lookup by cache handle
    struct PURGE_GET_FREE {};
    VListPtr(VNode_list& list, PURGE_GET_FREE);
  };

  friend class VListPtr;

  VNode_list(const VNode_list&);	// Unimplemented and forbidden: cloning
  void operator = (const VNode_list&);	// is not allowed

public:
  VNode_list(const int _size);
  VHandle open(const char name [], const int open_flags);
  FileStatus get_stat(const char name []);
  const char * get_cache_file(const VHandle handle) const;
  int get_cache_handle(const VHandle handle) const;
  bool close(const VHandle handle);
  bool close_cache_handle(const int cache_file_handle);
};

#endif
